@@ -30,6 +30,7 @@
 #include "util/slab.h"
 #include "util/u_debug.h"
 #include "util/macros.h"
+#include "util/u_framebuffer.h"
 #include "util/u_threaded_context.h"
 #include "intel/blorp/blorp.h"
 #include "intel/dev/intel_debug.h"
@@ -231,7 +232,8 @@ struct iris_vue_prog_key {
    struct iris_base_prog_key base;
 
    unsigned nr_userclip_plane_consts:4;
-   unsigned padding:28;
+   enum intel_vue_layout layout:2;
+   unsigned padding:26;
 };
 
 struct iris_vs_prog_key {
@@ -284,7 +286,8 @@ struct iris_fs_prog_key {
    bool multisample_fbo:1;
    bool force_dual_color_blend:1;
    bool coherent_fb_fetch:1;
-   uint64_t padding:43;
+   enum intel_vue_layout vue_layout:2;
+   uint64_t padding:41;
 };
 
 struct iris_cs_prog_key {
@@ -1006,6 +1009,7 @@ struct iris_context {
       struct pipe_viewport_state viewports[IRIS_MAX_VIEWPORTS];
       struct pipe_scissor_state scissors[IRIS_MAX_VIEWPORTS];
       struct pipe_stencil_ref stencil_ref;
+      PIPE_FB_SURFACES; //STOP USING THIS
       struct pipe_framebuffer_state framebuffer;
       struct pipe_clip_state clip_planes;
       /* width and height treated like x2 and y2 */
@@ -1102,7 +1106,7 @@ struct iris_context {
        * Query BO with a MI_PREDICATE_RESULT snapshot calculated on the
        * render context that needs to be uploaded to the compute context.
        */
-      struct iris_bo *compute_predicate;
+      struct iris_address compute_predicate;
 
       /** Is a PIPE_QUERY_PRIMITIVES_GENERATED query active? */
       bool prims_generated_query_active;
@@ -1155,6 +1159,11 @@ struct iris_context {
       struct pipe_resource *pixel_hashing_tables;
 
       bool use_tbimr;
+
+      /* For compute engine only */
+      uint8_t pixel_async_compute_thread_limit;
+      uint8_t z_pass_async_compute_thread_limit;
+      uint8_t np_z_async_throttle_settings;
    } state;
 };
 
@@ -1353,9 +1362,9 @@ iris_fs_barycentric_modes(const struct iris_compiled_shader *shader,
                           enum intel_msaa_flags pushed_msaa_flags);
 bool iris_use_tcs_multi_patch(struct iris_screen *screen);
 bool iris_indirect_ubos_use_sampler(struct iris_screen *screen);
-const void *iris_get_compiler_options(struct pipe_screen *pscreen,
-                                      enum pipe_shader_ir ir,
-                                      enum pipe_shader_type pstage);
+const struct nir_shader_compiler_options *
+iris_get_compiler_options(struct pipe_screen *pscreen,
+                          enum pipe_shader_type pstage);
 
 /* iris_disk_cache.c */
 
