@@ -17,8 +17,6 @@
 #include "util/u_video.h"
 #include "vl/vl_video_buffer.h"
 
-static const unsigned index_to_shifts[4] = {24, 16, 8, 0};
-
 /* set quality modes from the input */
 static void radeon_vcn_enc_quality_modes(struct radeon_encoder *enc,
                                          struct pipe_enc_quality_modes *in)
@@ -42,7 +40,7 @@ static void radeon_vcn_enc_quality_modes(struct radeon_encoder *enc,
 
    /* Disabling 2pass encoding for VCN 5.0
     * This is a temporary limitation only for VCN 5.0 due to HW,
-    * once verified in future VCN 5.X versions, it will be enabled again. 
+    * once verified in future VCN 5.X versions, it will be enabled again.
     */
    if (sscreen->info.vcn_ip_version >= VCN_5_0_0)
       p->pre_encode_mode = RENCODE_PREENCODE_MODE_NONE;
@@ -162,7 +160,8 @@ static void radeon_vcn_enc_get_roi_param(struct radeon_encoder *enc,
    else {
       uint32_t width_in_block, height_in_block;
       uint32_t block_length;
-      int32_t i, j, pa_format = 0;
+      int32_t i, j;
+      bool pa_format = false;
 
       qp_map->version = sscreen->info.vcn_ip_version >= VCN_5_0_0
                         ? RENCODE_QP_MAP_VCN5 : RENCODE_QP_MAP_LEGACY;
@@ -172,7 +171,7 @@ static void radeon_vcn_enc_get_roi_param(struct radeon_encoder *enc,
       if (enc->enc_pic.rc_session_init.rate_control_method &&
             (qp_map->version ==  RENCODE_QP_MAP_LEGACY)) {
          enc->enc_pic.enc_qp_map.qp_map_type = RENCODE_QP_MAP_TYPE_MAP_PA;
-         pa_format = 1;
+         pa_format = true;
       }
       else
          enc->enc_pic.enc_qp_map.qp_map_type = RENCODE_QP_MAP_TYPE_DELTA;
@@ -196,7 +195,7 @@ static void radeon_vcn_enc_get_roi_param(struct radeon_encoder *enc,
             /* mapped av1 qi into the legacy qp range by dividing by 5 and
              * rounding up in any rate control mode.
              */
-            if (is_av1 && (pa_format || (qp_map->version ==  RENCODE_QP_MAP_VCN5))) {
+            if (is_av1 && pa_format) {
                if (region->qp_value > 0)
                   av1_qi_value = (region->qp_value + 2) / 5;
                else if (region->qp_value < 0)
@@ -294,13 +293,18 @@ static void radeon_vcn_enc_h264_get_spec_misc_param(struct radeon_encoder *enc,
    enc->enc_pic.spec_misc.redundant_pic_cnt_present_flag =
       pic->pic_ctrl.redundant_pic_cnt_present_flag;
    enc->enc_pic.spec_misc.b_picture_enabled = !!pic->seq.max_num_reorder_frames;
+   /* FW issue when using constrained intra pred with rate control. */
    enc->enc_pic.spec_misc.constrained_intra_pred_flag =
+      !enc->enc_pic.rc_session_init.rate_control_method &&
       pic->pic_ctrl.constrained_intra_pred_flag;
    enc->enc_pic.spec_misc.half_pel_enabled = 1;
    enc->enc_pic.spec_misc.quarter_pel_enabled = 1;
-   enc->enc_pic.spec_misc.weighted_bipred_idc = 0;
+   enc->enc_pic.spec_misc.weighted_bipred_idc =
+      pic->pic_ctrl.weighted_bipred_idc != 1 ?
+      pic->pic_ctrl.weighted_bipred_idc : 0;
    enc->enc_pic.spec_misc.transform_8x8_mode =
       sscreen->info.vcn_ip_version >= VCN_5_0_0 &&
+      enc->enc_pic.spec_misc.cabac_enable &&
       pic->pic_ctrl.transform_8x8_mode_flag;
    enc->enc_pic.spec_misc.level_idc = pic->seq.level_idc;
 }
@@ -501,17 +505,45 @@ static void radeon_vcn_enc_h264_get_param(struct radeon_encoder *enc,
    enc->enc_pic.picture_type = pic->picture_type;
    enc->enc_pic.bit_depth_luma_minus8 = 0;
    enc->enc_pic.bit_depth_chroma_minus8 = 0;
-   enc->enc_pic.enc_params.reference_picture_index =
-      pic->ref_list0[0] == PIPE_H2645_LIST_REF_INVALID_ENTRY ? 0xffffffff : pic->ref_list0[0];
-   enc->enc_pic.h264_enc_params.l1_reference_picture0_index =
-      pic->ref_list1[0] == PIPE_H2645_LIST_REF_INVALID_ENTRY ? 0xffffffff : pic->ref_list1[0];
+   enc->enc_pic.h264_enc_params.input_pic_order_cnt = pic->pic_order_cnt;
    enc->enc_pic.h264_enc_params.input_picture_structure = RENCODE_H264_PICTURE_STRUCTURE_FRAME;
    enc->enc_pic.h264_enc_params.interlaced_mode = RENCODE_H264_INTERLACING_MODE_PROGRESSIVE;
    enc->enc_pic.h264_enc_params.l0_reference_picture1_index = 0xffffffff;
    enc->enc_pic.enc_params.reconstructed_picture_index = pic->dpb_curr_pic;
    enc->enc_pic.h264_enc_params.is_reference = !pic->not_referenced;
    enc->enc_pic.h264_enc_params.is_long_term = pic->is_ltr;
-   enc->enc_pic.not_referenced = pic->not_referenced;
+
+   if (pic->ref_list0[0] != PIPE_H2645_LIST_REF_INVALID_ENTRY) {
+      uint8_t ref_l0 = pic->ref_list0[0];
+
+      enc->enc_pic.enc_params.reference_picture_index = ref_l0;
+      enc->enc_pic.h264_enc_params.picture_info_l0_reference_picture0.pic_type =
+         radeon_enc_h2645_picture_type(pic->dpb[ref_l0].picture_type);
+      enc->enc_pic.h264_enc_params.picture_info_l0_reference_picture0.pic_order_cnt =
+         pic->dpb[ref_l0].pic_order_cnt;
+      enc->enc_pic.h264_enc_params.picture_info_l0_reference_picture0.is_long_term =
+         pic->dpb[ref_l0].is_ltr;
+   } else {
+      enc->enc_pic.enc_params.reference_picture_index = 0xffffffff;
+   }
+
+   if (pic->ref_list1[0] != PIPE_H2645_LIST_REF_INVALID_ENTRY) {
+      uint8_t ref_l1 = pic->ref_list1[0];
+
+      enc->enc_pic.h264_enc_params.l1_reference_picture0_index = ref_l1;
+      enc->enc_pic.h264_enc_params.picture_info_l1_reference_picture0.pic_type =
+         radeon_enc_h2645_picture_type(pic->dpb[ref_l1].picture_type);
+      enc->enc_pic.h264_enc_params.picture_info_l1_reference_picture0.pic_order_cnt =
+         pic->dpb[ref_l1].pic_order_cnt;
+      enc->enc_pic.h264_enc_params.picture_info_l1_reference_picture0.is_long_term =
+         pic->dpb[ref_l1].is_ltr;
+   } else {
+      enc->enc_pic.h264_enc_params.l1_reference_picture0_index = 0xffffffff;
+   }
+
+   if (pic->ref_list1[0] != PIPE_H2645_LIST_REF_INVALID_ENTRY &&
+       pic->dpb[pic->ref_list1[0]].picture_type == PIPE_H2645_ENC_PICTURE_TYPE_B)
+      RADEON_ENC_ERR("B-frame as L1 reference not supported\n");
 
    if (enc->dpb_type == DPB_TIER_2) {
       for (uint32_t i = 0; i < ARRAY_SIZE(pic->dpb); i++) {
@@ -595,6 +627,32 @@ static void radeon_vcn_enc_hevc_get_dbk_param(struct radeon_encoder *enc,
       !pic->seq.sample_adaptive_offset_enabled_flag;
 }
 
+static bool cu_qp_delta_supported(struct si_screen *sscreen)
+{
+   if (sscreen->info.vcn_ip_version >= VCN_5_0_0)
+      return true;
+   else if (sscreen->info.vcn_ip_version >= VCN_4_0_0)
+      return sscreen->info.vcn_enc_minor_version >= 7;
+   else if (sscreen->info.vcn_ip_version >= VCN_3_0_0)
+      return sscreen->info.vcn_enc_minor_version >= 26;
+   else if (sscreen->info.vcn_ip_version >= VCN_2_0_0)
+      return sscreen->info.vcn_enc_minor_version >= 20;
+   else
+      return false;
+}
+
+static bool transform_skip_supported(struct si_screen *sscreen)
+{
+   if (sscreen->info.vcn_ip_version >= VCN_5_0_0)
+      return true;
+   else if (sscreen->info.vcn_ip_version >= VCN_4_0_0)
+      return sscreen->info.vcn_enc_minor_version >= 2;
+   else if (sscreen->info.vcn_ip_version >= VCN_3_0_0)
+      return sscreen->info.vcn_enc_minor_version >= 23;
+   else
+      return false;
+}
+
 static void radeon_vcn_enc_hevc_get_spec_misc_param(struct radeon_encoder *enc,
                                                     struct pipe_h265_enc_picture_desc *pic)
 {
@@ -605,17 +663,18 @@ static void radeon_vcn_enc_hevc_get_spec_misc_param(struct radeon_encoder *enc,
    enc->enc_pic.hevc_spec_misc.amp_disabled = !pic->seq.amp_enabled_flag;
    enc->enc_pic.hevc_spec_misc.strong_intra_smoothing_enabled =
       pic->seq.strong_intra_smoothing_enabled_flag;
+   /* FW issue when using constrained intra pred with rate control. */
    enc->enc_pic.hevc_spec_misc.constrained_intra_pred_flag =
+      !enc->enc_pic.rc_session_init.rate_control_method &&
       pic->pic.constrained_intra_pred_flag;
    enc->enc_pic.hevc_spec_misc.cabac_init_flag = pic->slice.cabac_init_flag;
    enc->enc_pic.hevc_spec_misc.half_pel_enabled = 1;
    enc->enc_pic.hevc_spec_misc.quarter_pel_enabled = 1;
    enc->enc_pic.hevc_spec_misc.transform_skip_disabled =
-      sscreen->info.vcn_ip_version < VCN_3_0_0 ||
+      !transform_skip_supported(sscreen) ||
       !pic->pic.transform_skip_enabled_flag;
    enc->enc_pic.hevc_spec_misc.cu_qp_delta_enabled_flag =
-      (sscreen->info.vcn_ip_version >= VCN_2_0_0 &&
-      pic->pic.cu_qp_delta_enabled_flag) ||
+      (cu_qp_delta_supported(sscreen) && pic->pic.cu_qp_delta_enabled_flag) ||
       enc->enc_pic.enc_qp_map.qp_map_type ||
       enc->enc_pic.rc_session_init.rate_control_method;
 }
@@ -788,8 +847,6 @@ static void radeon_vcn_enc_av1_get_session_param(struct radeon_encoder *enc,
       enc->enc_pic.session_init.aligned_picture_height = align(height, align_height);
       if (!(height % 8) && (height % 16))
          enc->enc_pic.session_init.aligned_picture_height = height + 2;
-      enc->enc_pic.av1.coded_width = enc->enc_pic.session_init.aligned_picture_width;
-      enc->enc_pic.av1.coded_height = enc->enc_pic.session_init.aligned_picture_height;
       if (sscreen->info.vcn_ip_version == VCN_4_0_2 ||
           sscreen->info.vcn_ip_version == VCN_4_0_5 ||
           sscreen->info.vcn_ip_version == VCN_4_0_6)
@@ -799,9 +856,9 @@ static void radeon_vcn_enc_av1_get_session_param(struct radeon_encoder *enc,
       align_height = 2;
       enc->enc_pic.session_init.aligned_picture_width = align(width, align_width);
       enc->enc_pic.session_init.aligned_picture_height = align(height, align_height);
-      enc->enc_pic.av1.coded_width = width;
-      enc->enc_pic.av1.coded_height = height;
    }
+   enc->enc_pic.av1.coded_width = enc->enc_pic.session_init.aligned_picture_width;
+   enc->enc_pic.av1.coded_height = enc->enc_pic.session_init.aligned_picture_height;
 
    uint32_t padding_width = 0;
    uint32_t padding_height = 0;
@@ -975,13 +1032,10 @@ static void radeon_vcn_enc_av1_get_param(struct radeon_encoder *enc,
       pic->seq.bit_depth_minus8;
    enc_pic->pic_width_in_luma_samples = pic->seq.pic_width_in_luma_samples;
    enc_pic->pic_height_in_luma_samples = pic->seq.pic_height_in_luma_samples;
-   enc_pic->enable_render_size = pic->enable_render_size;
    enc_pic->enable_error_resilient_mode = pic->error_resilient_mode;
    enc_pic->force_integer_mv = pic->force_integer_mv;
    enc_pic->disable_screen_content_tools = !pic->allow_screen_content_tools;
    enc_pic->is_obu_frame = pic->enable_frame_obu;
-   enc_pic->render_width = pic->render_width;
-   enc_pic->render_height = pic->render_height;
 
    enc_pic->enc_params.reference_picture_index =
       pic->ref_list0[0] == PIPE_H2645_LIST_REF_INVALID_ENTRY ?
@@ -989,6 +1043,9 @@ static void radeon_vcn_enc_av1_get_param(struct radeon_encoder *enc,
    enc_pic->enc_params.reconstructed_picture_index = pic->dpb_curr_pic;
 
    if (sscreen->info.vcn_ip_version >= VCN_5_0_0) {
+      bool allow_unidir =
+         pic->rc[0].rate_ctrl_method == PIPE_H2645_ENC_RATE_CONTROL_METHOD_DISABLE;
+
       for (uint32_t i = 0; i < RENCODE_AV1_REFS_PER_FRAME; i++)
          enc_pic->av1_enc_params.ref_frames[i] = pic->dpb_ref_frame_idx[i];
 
@@ -1000,7 +1057,7 @@ static void radeon_vcn_enc_av1_get_param(struct radeon_encoder *enc,
       if (pic->ref_list1[0] != PIPE_H2645_LIST_REF_INVALID_ENTRY) {
          enc_pic->av1.compound = true; /* BIDIR_COMP */
          enc_pic->av1_enc_params.lsm_reference_frame_index[1] = pic->ref_list1[0];
-      } else if (pic->ref_list0[1] != PIPE_H2645_LIST_REF_INVALID_ENTRY) {
+      } else if (allow_unidir && pic->ref_list0[1] != PIPE_H2645_LIST_REF_INVALID_ENTRY) {
          enc_pic->av1.compound = true; /* UNIDIR_COMP */
          enc_pic->av1_enc_params.lsm_reference_frame_index[1] = pic->ref_list0[1];
       }
@@ -1085,11 +1142,6 @@ static void radeon_enc_flush(struct pipe_video_codec *encoder)
    flush(enc, PIPE_FLUSH_ASYNC, NULL);
 }
 
-static void radeon_enc_cs_flush(void *ctx, unsigned flags, struct pipe_fence_handle **fence)
-{
-   // just ignored
-}
-
 /* configure reconstructed picture offset */
 static void radeon_enc_rec_offset(rvcn_enc_reconstructed_picture_t *recon,
                                   uint32_t *offset,
@@ -1171,7 +1223,7 @@ static int setup_cdf(struct radeon_encoder *enc)
 
    p_cdf = enc->ws->buffer_map(enc->ws,
                                enc->cdf->res->buf,
-                              &enc->cs,
+                               NULL,
                                PIPE_MAP_READ_WRITE | RADEON_MAP_TEMPORARY);
    if (!p_cdf)
       goto error;
@@ -1188,6 +1240,7 @@ error:
 static void pre_encode_size(struct radeon_encoder *enc,
                             uint32_t *offset)
 {
+   struct si_screen *sscreen = (struct si_screen *)enc->screen;
    bool is_h264 = u_reduce_video_profile(enc->base.profile)
                              == PIPE_VIDEO_FORMAT_MPEG4_AVC;
    uint32_t rec_alignment = is_h264 ? 16 : 64;
@@ -1201,10 +1254,18 @@ static void pre_encode_size(struct radeon_encoder *enc,
       DIV_ROUND_UP(aligned_height, rec_alignment);
 
    enc_pic->ctx_buf.two_pass_search_center_map_offset = *offset;
-   if (is_h264 && !has_b)
-      *offset += align((pre_size * 4 + full_size) * sizeof(uint32_t), enc->alignment);
-   else if (!is_h264)
-      *offset += align((pre_size * 52 + full_size) * sizeof(uint32_t), enc->alignment);
+
+   if (sscreen->info.vcn_ip_version < VCN_5_0_0) {
+      if (is_h264 && !has_b)
+         *offset += align((pre_size * 4 + full_size) * sizeof(uint32_t), enc->alignment);
+      else if (!is_h264)
+         *offset += align((pre_size * 52 + full_size) * sizeof(uint32_t), enc->alignment);
+   } else { /* only for vcn5.x rather than VCN5_0_0 */
+      if (is_h264 && !has_b)
+         *offset += align(full_size * 8, enc->alignment);
+      else if (!is_h264)
+         *offset += align(full_size * 24, enc->alignment);
+   }
 }
 
 static int setup_dpb(struct radeon_encoder *enc, uint32_t num_reconstructed_pictures)
@@ -1238,8 +1299,6 @@ static int setup_dpb(struct radeon_encoder *enc, uint32_t num_reconstructed_pict
    enc_pic->ctx_buf.rec_luma_pitch   = pitch;
    enc_pic->ctx_buf.pre_encode_picture_luma_pitch   = pitch;
    enc_pic->ctx_buf.num_reconstructed_pictures = num_reconstructed_pictures;
-   enc_pic->dpb_luma_size   = luma_size;
-   enc_pic->dpb_chroma_size = chroma_size;
    enc_pic->total_coloc_bytes = total_coloc_bytes;
 
    offset = 0;
@@ -1412,7 +1471,7 @@ static int generate_roi_map(struct radeon_encoder *enc)
 
    p_roi = enc->ws->buffer_map(enc->ws,
                                enc->roi->res->buf,
-                              &enc->cs,
+                               NULL,
                                PIPE_MAP_READ_WRITE | RADEON_MAP_TEMPORARY);
    if (!p_roi)
       goto error;
@@ -1521,12 +1580,12 @@ static void radeon_enc_begin_frame(struct pipe_video_codec *encoder,
 
    if (dpb_slots > enc->dpb_slots) {
       setup_dpb(enc, dpb_slots);
-      if (!si_vid_resize_buffer(enc->base.context, &enc->cs, enc->dpb, enc->dpb_size, NULL)) {
+      if (!si_vid_resize_buffer(enc->base.context, enc->dpb, enc->dpb_size, NULL)) {
          RADEON_ENC_ERR("Can't resize DPB buffer.\n");
          goto error;
       }
       if (sscreen->info.vcn_ip_version >= VCN_5_0_0 && enc->metadata_size &&
-          !si_vid_resize_buffer(enc->base.context, &enc->cs, enc->meta, enc->metadata_size, NULL)) {
+          !si_vid_resize_buffer(enc->base.context, enc->meta, enc->metadata_size, NULL)) {
          RADEON_ENC_ERR("Can't resize meta buffer.\n");
          goto error;
       }
@@ -1562,12 +1621,10 @@ static void radeon_enc_begin_frame(struct pipe_video_codec *encoder,
 
    enc->need_feedback = false;
 
-   if (!enc->stream_handle) {
+   if (!enc->si) {
       struct rvid_buffer fb;
-      enc->stream_handle = si_vid_alloc_stream_handle();
       enc->si = CALLOC_STRUCT(rvid_buffer);
       if (!enc->si ||
-          !enc->stream_handle ||
           !si_vid_create_buffer(enc->screen, enc->si, 128 * 1024, PIPE_USAGE_DEFAULT)) {
          RADEON_ENC_ERR("Can't create session buffer.\n");
          goto error;
@@ -1675,7 +1732,7 @@ static void *radeon_vcn_enc_encode_headers(struct radeon_encoder *enc)
    if (!data)
       return NULL;
 
-   uint8_t *ptr = enc->ws->buffer_map(enc->ws, enc->bs_handle, &enc->cs,
+   uint8_t *ptr = enc->ws->buffer_map(enc->ws, enc->bs_handle, NULL,
                                       PIPE_MAP_WRITE | RADEON_MAP_TEMPORARY);
    if (!ptr) {
       RADEON_ENC_ERR("Can't map bs buffer.\n");
@@ -1764,14 +1821,14 @@ static int radeon_enc_end_frame(struct pipe_video_codec *encoder, struct pipe_vi
    if (enc->error)
       return -1;
 
-   return flush(enc, picture->flush_flags, picture->fence);
+   return flush(enc, picture->flush_flags, picture->out_fence);
 }
 
 static void radeon_enc_destroy(struct pipe_video_codec *encoder)
 {
    struct radeon_encoder *enc = (struct radeon_encoder *)encoder;
 
-   if (enc->stream_handle) {
+   if (enc->si) {
       struct rvid_buffer fb;
       enc->need_feedback = false;
       si_vid_create_buffer(enc->screen, &fb, 512, PIPE_USAGE_STAGING);
@@ -1799,7 +1856,7 @@ static void radeon_enc_get_feedback(struct pipe_video_codec *encoder, void *feed
    struct radeon_encoder *enc = (struct radeon_encoder *)encoder;
    struct rvid_buffer *fb = feedback;
 
-   uint32_t *ptr = enc->ws->buffer_map(enc->ws, fb->res->buf, &enc->cs,
+   uint32_t *ptr = enc->ws->buffer_map(enc->ws, fb->res->buf, NULL,
                                        PIPE_MAP_READ_WRITE | RADEON_MAP_TEMPORARY);
    if (ptr[1])
       *size = ptr[6] - ptr[8];
@@ -1887,7 +1944,6 @@ void radeon_enc_create_dpb_aux_buffers(struct radeon_encoder *enc, struct radeon
       return;
 
    uint32_t fcb_size = radeon_enc_frame_context_buffer_size(enc);
-   uint32_t recon_size = enc->enc_pic.dpb_luma_size + enc->enc_pic.dpb_chroma_size;
 
    buf->fcb = CALLOC_STRUCT(rvid_buffer);
    if (!buf->fcb || !si_vid_create_buffer(enc->screen, buf->fcb, fcb_size, PIPE_USAGE_DEFAULT)) {
@@ -1896,11 +1952,13 @@ void radeon_enc_create_dpb_aux_buffers(struct radeon_encoder *enc, struct radeon
    }
 
    if (enc->enc_pic.quality_modes.pre_encode_mode) {
-      buf->pre = CALLOC_STRUCT(rvid_buffer);
-      if (!buf->pre || !si_vid_create_buffer(enc->screen, buf->pre, recon_size, PIPE_USAGE_DEFAULT)) {
+      buf->pre = enc->base.context->create_video_buffer(enc->base.context, &buf->templ);
+      if (!buf->pre) {
          RADEON_ENC_ERR("Can't create preenc buffer!\n");
          return;
       }
+      buf->pre_luma = (struct si_texture *)((struct vl_video_buffer *)buf->pre)->resources[0];
+      buf->pre_chroma = (struct si_texture *)((struct vl_video_buffer *)buf->pre)->resources[1];
 
       buf->pre_fcb = CALLOC_STRUCT(rvid_buffer);
       if (!buf->pre_fcb || !si_vid_create_buffer(enc->screen, buf->pre_fcb, fcb_size, PIPE_USAGE_DEFAULT)) {
@@ -1914,8 +1972,10 @@ static void radeon_enc_destroy_dpb_buffer(void *data)
 {
    struct radeon_enc_dpb_buffer *dpb = data;
 
+   if (dpb->pre)
+      dpb->pre->destroy(dpb->pre);
+
    RADEON_ENC_DESTROY_VIDEO_BUFFER(dpb->fcb);
-   RADEON_ENC_DESTROY_VIDEO_BUFFER(dpb->pre);
    RADEON_ENC_DESTROY_VIDEO_BUFFER(dpb->pre_fcb);
    FREE(dpb);
 }
@@ -1926,13 +1986,16 @@ static struct pipe_video_buffer *radeon_enc_create_dpb_buffer(struct pipe_video_
 {
    struct radeon_encoder *enc = (struct radeon_encoder *)encoder;
 
-   struct pipe_video_buffer *buf = enc->base.context->create_video_buffer(enc->base.context, templat);
+   struct pipe_video_buffer templ = *templat;
+   templ.bind |= PIPE_BIND_VIDEO_ENCODE_DPB;
+   struct pipe_video_buffer *buf = enc->base.context->create_video_buffer(enc->base.context, &templ);
    if (!buf) {
       RADEON_ENC_ERR("Can't create dpb buffer!\n");
       return NULL;
    }
 
    struct radeon_enc_dpb_buffer *dpb = CALLOC_STRUCT(radeon_enc_dpb_buffer);
+   dpb->templ = templ;
    dpb->luma = (struct si_texture *)((struct vl_video_buffer *)buf)->resources[0];
    dpb->chroma = (struct si_texture *)((struct vl_video_buffer *)buf)->resources[1];
 
@@ -1956,7 +2019,7 @@ struct pipe_video_codec *radeon_create_encoder(struct pipe_context *context,
       return NULL;
 
    if (sctx->vcn_has_ctx) {
-      enc->ectx = pipe_create_multimedia_context(context->screen);
+      enc->ectx = context->screen->context_create(context->screen, NULL, PIPE_CONTEXT_COMPUTE_ONLY);
       if (!enc->ectx)
          sctx->vcn_has_ctx = false;
    }
@@ -1973,13 +2036,12 @@ struct pipe_video_codec *radeon_create_encoder(struct pipe_context *context,
    enc->base.fence_wait = radeon_enc_fence_wait;
    enc->base.destroy_fence = radeon_enc_destroy_fence;
    enc->get_buffer = get_buffer;
-   enc->bits_in_shifter = 0;
    enc->screen = context->screen;
    enc->ws = ws;
 
    if (!ws->cs_create(&enc->cs,
        (sctx->vcn_has_ctx) ? ((struct si_context *)enc->ectx)->ctx : sctx->ctx,
-       AMD_IP_VCN_ENC, radeon_enc_cs_flush, enc)) {
+       AMD_IP_VCN_ENC, NULL, NULL)) {
       RADEON_ENC_ERR("Can't get command submission context.\n");
       goto error;
    }
@@ -2000,14 +2062,18 @@ struct pipe_video_codec *radeon_create_encoder(struct pipe_context *context,
          /* this limits tile splitting scheme to use legacy method */
          enc->enc_pic.av1_tile_splitting_legacy_flag = true;
       }
+      if (sscreen->info.vcn_enc_minor_version >= 8)
+         enc->enc_pic.has_dependent_slice_instructions = true;
    }
    else if (sscreen->info.vcn_ip_version >= VCN_4_0_0) {
       if (sscreen->info.vcn_enc_minor_version >= 1)
          enc->enc_pic.use_rc_per_pic_ex = true;
+      if (sscreen->info.vcn_enc_minor_version >= 23)
+         enc->enc_pic.has_dependent_slice_instructions = true;
       radeon_enc_4_0_init(enc);
    }
    else if (sscreen->info.vcn_ip_version >= VCN_3_0_0) {
-      if (sscreen->info.vcn_enc_minor_version >= 24)
+      if (sscreen->info.vcn_enc_minor_version >= 29)
          enc->enc_pic.use_rc_per_pic_ex = true;
       radeon_enc_3_0_init(enc);
    }
@@ -2041,96 +2107,6 @@ void radeon_enc_add_buffer(struct radeon_encoder *enc, struct pb_buffer_lean *bu
    RADEON_ENC_CS(addr);
 }
 
-void radeon_enc_set_emulation_prevention(struct radeon_encoder *enc, bool set)
-{
-   if (set != enc->emulation_prevention) {
-      enc->emulation_prevention = set;
-      enc->num_zeros = 0;
-   }
-}
-
-void radeon_enc_set_output_buffer(struct radeon_encoder *enc, uint8_t *buffer)
-{
-   enc->bits_buf = buffer;
-   enc->bits_buf_pos = 0;
-}
-
-void radeon_enc_output_one_byte(struct radeon_encoder *enc, unsigned char byte)
-{
-   if (enc->bits_buf) {
-      enc->bits_buf[enc->bits_buf_pos++] = byte;
-      return;
-   }
-
-   if (enc->byte_index == 0)
-      enc->cs.current.buf[enc->cs.current.cdw] = 0;
-   enc->cs.current.buf[enc->cs.current.cdw] |=
-      ((unsigned int)(byte) << index_to_shifts[enc->byte_index]);
-   enc->byte_index++;
-
-   if (enc->byte_index >= 4) {
-      enc->byte_index = 0;
-      enc->cs.current.cdw++;
-   }
-}
-
-void radeon_enc_emulation_prevention(struct radeon_encoder *enc, unsigned char byte)
-{
-   if (enc->emulation_prevention) {
-      if ((enc->num_zeros >= 2) && ((byte == 0x00) || (byte == 0x01) ||
-         (byte == 0x02) || (byte == 0x03))) {
-         radeon_enc_output_one_byte(enc, 0x03);
-         enc->bits_output += 8;
-         enc->num_zeros = 0;
-      }
-      enc->num_zeros = (byte == 0 ? (enc->num_zeros + 1) : 0);
-   }
-}
-
-void radeon_enc_code_fixed_bits(struct radeon_encoder *enc, unsigned int value,
-                                unsigned int num_bits)
-{
-   unsigned int bits_to_pack = 0;
-   enc->bits_size += num_bits;
-
-   while (num_bits > 0) {
-      unsigned int value_to_pack = value & (0xffffffff >> (32 - num_bits));
-      bits_to_pack =
-         num_bits > (32 - enc->bits_in_shifter) ? (32 - enc->bits_in_shifter) : num_bits;
-
-      if (bits_to_pack < num_bits)
-         value_to_pack = value_to_pack >> (num_bits - bits_to_pack);
-
-      enc->shifter |= value_to_pack << (32 - enc->bits_in_shifter - bits_to_pack);
-      num_bits -= bits_to_pack;
-      enc->bits_in_shifter += bits_to_pack;
-
-      while (enc->bits_in_shifter >= 8) {
-         unsigned char output_byte = (unsigned char)(enc->shifter >> 24);
-         enc->shifter <<= 8;
-         radeon_enc_emulation_prevention(enc, output_byte);
-         radeon_enc_output_one_byte(enc, output_byte);
-         enc->bits_in_shifter -= 8;
-         enc->bits_output += 8;
-      }
-   }
-}
-
-void radeon_enc_code_uvlc(struct radeon_encoder *enc, unsigned int value)
-{
-   uint32_t num_bits = 0;
-   uint64_t value_plus1 = (uint64_t)value + 1;
-   uint32_t num_leading_zeros = 0;
-
-   while ((uint64_t)1 << num_bits <= value_plus1)
-      num_bits++;
-
-   num_leading_zeros = num_bits - 1;
-   radeon_enc_code_fixed_bits(enc, 0, num_leading_zeros);
-   radeon_enc_code_fixed_bits(enc, 1, 1);
-   radeon_enc_code_fixed_bits(enc, (uint32_t)value_plus1, num_leading_zeros);
-}
-
 void radeon_enc_code_leb128(uint8_t *buf, uint32_t value,
                             uint32_t num_bytes)
 {
@@ -2149,71 +2125,6 @@ void radeon_enc_code_leb128(uint8_t *buf, uint32_t value,
    } while((leb128_byte & 0x80));
 }
 
-void radeon_enc_reset(struct radeon_encoder *enc)
-{
-   enc->emulation_prevention = false;
-   enc->shifter = 0;
-   enc->bits_in_shifter = 0;
-   enc->bits_output = 0;
-   enc->num_zeros = 0;
-   enc->byte_index = 0;
-   enc->bits_size = 0;
-   enc->bits_buf = NULL;
-   enc->bits_buf_pos = 0;
-}
-
-void radeon_enc_byte_align(struct radeon_encoder *enc)
-{
-   unsigned int num_padding_zeros = (32 - enc->bits_in_shifter) % 8;
-
-   if (num_padding_zeros > 0)
-      radeon_enc_code_fixed_bits(enc, 0, num_padding_zeros);
-}
-
-void radeon_enc_flush_headers(struct radeon_encoder *enc)
-{
-   if (enc->bits_in_shifter != 0) {
-      unsigned char output_byte = (unsigned char)(enc->shifter >> 24);
-      radeon_enc_emulation_prevention(enc, output_byte);
-      radeon_enc_output_one_byte(enc, output_byte);
-      enc->bits_output += enc->bits_in_shifter;
-      enc->shifter = 0;
-      enc->bits_in_shifter = 0;
-      enc->num_zeros = 0;
-   }
-
-   if (enc->byte_index > 0) {
-      enc->cs.current.cdw++;
-      enc->byte_index = 0;
-   }
-}
-
-void radeon_enc_code_ue(struct radeon_encoder *enc, unsigned int value)
-{
-   unsigned int x = 0;
-   unsigned int ue_code = value + 1;
-   value += 1;
-
-   while (value) {
-      value = (value >> 1);
-      x += 1;
-   }
-
-   if (x > 1)
-     radeon_enc_code_fixed_bits(enc, 0, x - 1);
-   radeon_enc_code_fixed_bits(enc, ue_code, x);
-}
-
-void radeon_enc_code_se(struct radeon_encoder *enc, int value)
-{
-   unsigned int v = 0;
-
-   if (value != 0)
-      v = (value < 0 ? ((unsigned int)(0 - value) << 1) : (((unsigned int)(value) << 1) - 1));
-
-   radeon_enc_code_ue(enc, v);
-}
-
 unsigned int radeon_enc_av1_tile_log2(unsigned int blk_size, unsigned int max)
 {
    unsigned int k;
@@ -2224,27 +2135,36 @@ unsigned int radeon_enc_av1_tile_log2(unsigned int blk_size, unsigned int max)
    return k;
 }
 
-void radeon_enc_code_ns(struct radeon_encoder *enc, unsigned int value, unsigned int max)
+unsigned int radeon_enc_h2645_picture_type(enum pipe_h2645_enc_picture_type type)
 {
-   unsigned w = 0;
-   unsigned m;
-   unsigned max_num = max;
-
-   while ( max_num ) {
-      max_num >>= 1;
-      w++;
+   switch (type) {
+   case PIPE_H2645_ENC_PICTURE_TYPE_I:
+   case PIPE_H2645_ENC_PICTURE_TYPE_IDR:
+      return RENCODE_PICTURE_TYPE_I;
+   case PIPE_H2645_ENC_PICTURE_TYPE_P:
+      return RENCODE_PICTURE_TYPE_P;
+   case PIPE_H2645_ENC_PICTURE_TYPE_SKIP:
+      return RENCODE_PICTURE_TYPE_P_SKIP;
+   case PIPE_H2645_ENC_PICTURE_TYPE_B:
+      return RENCODE_PICTURE_TYPE_B;
+   default:
+      assert(false);
+      return 0;
    }
+}
 
-   m = ( 1 << w ) - max;
-
-   assert(w > 1);
-
-   if ( value < m )
-      radeon_enc_code_fixed_bits(enc, value, (w - 1));
-   else {
-      unsigned diff = value - m;
-      unsigned out = (((diff >> 1) + m) << 1) | (diff & 0x1);
-      radeon_enc_code_fixed_bits(enc, out, w);
+unsigned int radeon_enc_av1_picture_type(enum pipe_av1_enc_frame_type type)
+{
+   switch (type) {
+   case PIPE_AV1_ENC_FRAME_TYPE_KEY:
+   case PIPE_AV1_ENC_FRAME_TYPE_INTRA_ONLY:
+      return RENCODE_PICTURE_TYPE_I;
+   case PIPE_AV1_ENC_FRAME_TYPE_INTER:
+   case PIPE_AV1_ENC_FRAME_TYPE_SWITCH:
+      return RENCODE_PICTURE_TYPE_P;
+   default:
+      assert(false);
+      return 0;
    }
 }
 
@@ -2262,13 +2182,14 @@ static void radeon_enc_av1_bs_copy_end(struct radeon_encoder *enc, uint32_t bits
 
 /* av1 bitstream instruction type */
 void radeon_enc_av1_bs_instruction_type(struct radeon_encoder *enc,
+                                        struct radeon_bitstream *bs,
                                         uint32_t inst,
                                         uint32_t obu_type)
 {
-   radeon_enc_flush_headers(enc);
+   radeon_bs_flush_headers(bs);
 
-   if (enc->bits_output)
-      radeon_enc_av1_bs_copy_end(enc, enc->bits_output);
+   if (bs->bits_output)
+      radeon_enc_av1_bs_copy_end(enc, bs->bits_output);
 
    enc->enc_pic.copy_start = &enc->cs.current.buf[enc->cs.current.cdw++];
    RADEON_ENC_CS(inst);
@@ -2282,7 +2203,7 @@ void radeon_enc_av1_bs_instruction_type(struct radeon_encoder *enc,
    } else
       RADEON_ENC_CS(0); /* allocate a dword for number of bits */
 
-   radeon_enc_reset(enc);
+   radeon_bs_reset(bs, NULL, &enc->cs);
 }
 
 uint32_t radeon_enc_value_bits(uint32_t value)
